import ChannelManager from './channelManager'

const logger = require('./logger');
const url = require('url');
const http = require('http');
const crypto = require('crypto');
const app = require('./config/app');
const cors = require('cors');

const AUTH_TOKEN = app.archimedesToken;
const HTTP_PORT = app.port;

import { Server, IncomingMessage, ServerResponse } from 'http'

interface ArchimedesMessage {
  examId: string,
  reservation_no: string
}

class XhrListener {
  #cm: ChannelManager;

  constructor(cm: ChannelManager, server: any, app: any, jsonParser: any) {

    this.#cm = cm;
    logger.debugLog.info("XhrListener running");

    app.use(jsonParser);

    // TODO: try to pin down origin to reflect client rather than use wildcard
    app.options('*', cors());

    // I'm not entirely happy with bind(this) but it does let me break the
    // callback and event handlers out into class methods. It will be
    // problematic if the actual `this` needs to be accessed.

    app.post('/event/resume', cors(),
      (req: any, res: any, next: any) => this.eventFromArchimedes.bind(this)("resumeExam", req, res));
    app.post('/event/pause', cors(),
      (req: any, res: any, next: any) => this.eventFromArchimedes.bind(this)("pauseExam", req, res));

    server.on('request', app);

    // note that this kicks off HTTP listen for the entire app - not just HTTP but WS as well!
    server.listen(HTTP_PORT, () => logger.debugLog.info(`braid is now listening for HTTP and WS connections on port ${HTTP_PORT}`));
  }

  public authorized(authOffered: String): boolean {
    return AUTH_TOKEN === authOffered;
  }

  public eventFromArchimedes(eventType: string, req: any, res: any) {
    logger.debugLog.info(`eventFromArchimedes("${eventType}") called`);
    let authorized: boolean = this.authorized(req.headers['x-proctoru-signature']);

    if (!authorized) {
      res.status(401);
      res.send("Unauthorized");
      return;
    }

    res.setHeader('Content-Type', 'application/json');
    let amsg: ArchimedesMessage = { 'examId': req.body.examId, 'reservation_no': req.body.reservation_no };
    if (this.relayEvent(eventType, amsg)) {
      logger.debugLog.info('200 success relaying ' + eventType);
      res.send(JSON.stringify(`${eventType} event was successfully relayed`));
    } else {
      logger.errorLog.info(`XhrListener: could not relay ${eventType} event "${amsg}" to Exam UI; body: "${req.body}"`);
      res.status(400);
      res.send(JSON.stringify(`error while relaying ${eventType} event`));
    }
  }

  public relayEvent(event: string, archMsg: ArchimedesMessage): boolean {
    logger.debugLog.info(`XhrListener:relayEvent(event: ${event}, channel: ${archMsg.reservation_no})`);
    for (let c of this.#cm.channels) {
      if (archMsg.reservation_no == c.id) {
        let dmCount: number = 0;
        for (let client of c.clients) {
          let nonce: string = crypto.randomBytes(4).toString("hex");
          // TODO: verify the nonce against the received reply, which we currently ignore
          client.directMessage({
            message_type: "broadcast",
            message: { event_type: event, seq_id: nonce, examId: archMsg.examId }
          });
          dmCount++;
        }
        // dmCount of 1 would be normal. more than 1 is odd, but not necessarily bad. 0 means Exam UI has gone away somehow.
        return dmCount > 0;
      }
    }
    // did not match a channel; examId isn't valid for some reason
    return false;
  }
}

export default XhrListener;
